(* 3D Eulerian incompressible constant density fluid *)

type hidden_type =
{
  mutable c1: Grid.t;
  mutable c2: Grid.t;
  mutable c3: Grid.t;
  mutable u1: Grid.t;
  mutable v1: Grid.t;
}

type t =
{
  nx: int;
  ny: int;
  h: float;
  density: float;
  mutable u: Grid.t;
  mutable v: Grid.t;
  mutable p: Grid.t;
  occupied: Intgrid.t;
  hidden: hidden_type;
}


(* parameters for solving pressure *)
let maxits = 10
and tol = 1.e-3


(* initialize fluid at rest with given dimensions, density *)
let make nx ny h density =
  let hidden =
    {
      c1 = Grid.make nx ny;
      c2 = Grid.make nx ny;
      c3 = Grid.make nx ny;
      u1 = Grid.make (nx+1) ny;
      v1 = Grid.make nx (ny+1);
    } in
  {
    nx = nx;
    ny = ny;
    h = h;
    density = density;
    u = Grid.make (nx+1) ny;
    v = Grid.make nx (ny+1);
    p = Grid.make nx ny;
    occupied = Intgrid.make nx ny;
    hidden = hidden;
  }


(* Trace a particle backwards in velocity field, given initial position (in
 * grid, not real space), forward velocity, and ds (dt scaled for grid).
 * Return the coordinates at the end *)
let trace_backwards x y u v f ds =
  x-.ds*.u, y-.ds*.v


(* Track velocities back in field, add external force *)
let semi_lagrangian f dt =
  (* new cell-centred velocities, scaled by dt/h *)
  let ds = dt/.f.h in
  let cu = f.hidden.c1
  and cv = f.hidden.c2 in
  for i = 0 to f.nx-1 do
    for j = 0 to f.ny-1 do
      cu.{i,j} <- 0.5 *. (f.u.{i,j} +. f.u.{i+1,j});
      cv.{i,j} <- 0.5 *. (f.v.{i,j} +. f.v.{i,j+1});
  done done;
  (* now apply semi-lagrangian technique to each of u, v, w separately *)
  let maxx = float_of_int f.nx
  and maxy = float_of_int f.ny in
  (* first u *)
  let newu = f.hidden.u1 in
  for i = 0 to f.nx do
    let x = float_of_int i in
    for j = 0 to f.ny-1 do
      (* check if this doesn't border an immersed object *)
      if (i == f.nx || f.occupied.{i,j} == 0) &
         (i == 0 || f.occupied.{i-1,j} == 0) then begin
        let y = 0.5 +. float_of_int j in
        let u = f.u.{i,j}
        and v = 0.5*.( (if i > 0    then cv.{i-1,j} else cv.{i,j}) +.
                       (if i < f.nx then cv.{i,j}   else cv.{i-1,j}) ) in
        let ox,oy = trace_backwards x y u v f ds in
        let offy = oy -. 0.5 in   (* offset to match u grid *)
        (* check if a boundary condition applies *)
        if ox < 0.0 || ox >= maxx || offy < 0.0 || offy >= maxy-.1.000001 then
          newu.{i,j} <- 0.0
        else
          (* interpolate old velocity at that point to get new velocity *)
          newu.{i,j} <- Grid.interp f.u ox offy
      end
  done done;
  (* next v *)
  let newv = f.hidden.v1 in
  for i = 0 to f.nx-1 do
    let x = 0.5 +. float_of_int i in
    for j = 0 to f.ny do
      (* check if this doesn't border an immersed object *)
      if (j == f.ny || f.occupied.{i,j} == 0) &
         (j == 0 || f.occupied.{i,j-1} == 0) then begin
        let y = float_of_int j in
        let u = 0.5*.( (if j > 0    then cu.{i,j-1} else cu.{i,j}) +.
                       (if j < f.ny then cu.{i,j}   else cu.{i,j-1}) )
        and v = f.v.{i,j} in
        let ox,oy = trace_backwards x y u v f ds in
        let offx = ox -. 0.5 in
        (* check if a boundary condition applies *)
        if offx < 0.0 || offx >= maxx-.1.0000001 || oy < 0.0 || oy >= maxy then
          newv.{i,j} <- 0.0
        else
          (* interpolate old velocity at that point to get new velocity *)
          newv.{i,j} <- Grid.interp f.v offx oy
      end
  done done;
  (* and now swap old and new velocities *)
  f.hidden.u1 <- f.u;
  f.u <- newu;
  f.hidden.v1 <- f.v;
  f.v <- newv


(* add a random perturbation to velocity field of fluid *)
let perturb_fluid f perturb =
  for i = 0 to f.nx-1 do
    for j = 0 to f.ny-1 do
      f.u.{i,j} <- f.u.{i,j} +. perturb*.0.5 -. Random.float perturb;
      f.v.{i,j} <- f.v.{i,j} +. perturb*.0.5 -. Random.float perturb;
  done done


(* full timestep for fluid advancement *)
let advance f dt perturb =
  (* do convection term *)
  semi_lagrangian f dt;
  (* perturb fluid *)
  if perturb > 0.0 then
    perturb_fluid f perturb
  else ();
  (* solve for pressure to make fluid incompressible again *)
  let div = f.hidden.c1 in
  Grid.divergence ~x:f.u ~y:f.v ~div;
  Grid.rescale f.p (dt/.f.h);   (* to restore a good initial guess *)
  Grid.solve_poisson f.occupied ~lhs:f.p ~rhs:div f.hidden.c2 f.hidden.c3
    maxits tol;
  (* apply the pressure term *)
  Grid.sub_gradient f.occupied f.p ~x:f.u ~y:f.v;
  (* and finally rescale the pressure to physically correct dimension *)
  Grid.rescale f.p (f.h/.dt)


(* These are inconsistent with particle rendering for nonsquare grids or
 * graphics windows... *)

(* show velocity field *)
let render f =
  let m = (float (Graphics.size_x ()))/.(float (max f.nx f.ny)) in
  for i = 0 to f.nx-1 do
    let x = int_of_float (m *. (0.5 +. float i)) in
    for j = 0 to f.ny-1 do
      let y = int_of_float (m *. (0.5 +. float j)) in
      let cu = 0.5 *. (f.u.{i,j} +. f.u.{i+1,j})
      and cv = 0.5 *. (f.v.{i,j} +. f.v.{i,j+1}) in
      let rescale = 0.75 *. m /. (0.2 +. max (abs_float cu) (abs_float cv)) in
      let dx = int_of_float (cu*.rescale)
      and dy = int_of_float (cv*.rescale) in
      Graphics.moveto x y;
      Graphics.lineto (x+dx) (y+dy);
    done
  done

(* show pressure *)
let render_pressure f =
  let m = (float (Graphics.size_x ()))/.(float (max f.nx f.ny)) in
  let width = int_of_float m in
  for i = 0 to f.nx-1 do
    let x = int_of_float (m *. float i) in
    for j = 0 to f.ny-1 do
      let y = int_of_float (m *. float j) in
      let c = min 255 (max 0 (int_of_float (15.0*.f.p.{i,j} +. 127.0))) in
      Graphics.set_color c;
      Graphics.fill_rect x y width width;
    done
  done

