(* grid operations in service of Fluid module *)

open Bigarray

type t = (float, float64_elt, c_layout) Array2.t


(* create a grid with given dimensions, initially zero *)
let make nx ny =
  let g = Array2.create float64 c_layout nx ny in
  for i = 0 to nx-1 do
    for j = 0 to ny-1 do
      g.{i,j} <- 0.0
  done done;
  g


(* replace given grid with scalar times grid *)
let rescale (g:t) a =
  let n = (Array2.dim1 g)*(Array2.dim2 g) in
  let g1 = reshape_1 (genarray_of_array2 g) n in
  for i = 0 to n-1 do
    g1.{i} <- a*.g1.{i}
  done


(* trilinear interpolation between integer indices:
 * x, y, and z must be in the range [0,maxdim-1) for the appropriate maxdim. *)
let interp (g:t) x y =
  let i = int_of_float x
  and j = int_of_float y in
  let xfrac = x -. (float_of_int i)
  and yfrac = y -. (float_of_int j) in
  let othery = 1.0-.yfrac in
  (1.0-.xfrac) *. (othery*.g.{i,j} +. yfrac*.g.{i,j+1})
      +. xfrac *. (othery*.g.{i+1,j} +. yfrac*.g.{i+1,j+1})


(* Find the (undivided - no delta x etc.) divergence of the first grids,
 * which we assume are staggered, and store it in the last (cell-centred).
 * Uses second order central differences. *)
let divergence ~(x:t) ~(y:t) ~(div:t) =
  let endi = (Array2.dim1 x)-2
  and endj = (Array2.dim2 y)-2 in
  for i = 0 to endi do
    for j = 0 to endj do
      let dx = x.{i+1,j}-.x.{i,j}
      and dy = y.{i,j+1}-.y.{i,j} in
      div.{i,j} <- dx +. dy
  done done


(* Subtract (undivided) gradient of first grid to the other three grids.
 * Uses first order forward differences. Assume Dirichlet outer boundary,
 * but Neumann at boundaries of immersed objects. *)
let sub_gradient (occupied:Intgrid.t) (g:t) ~(x:t) ~(y:t) =
  let ni = Array2.dim1 g
  and nj = Array2.dim2 g in
  for i = 0 to ni do
    for j = 0 to nj-1 do
      if (i == ni || occupied.{i,j} == 0) &
         (i == 0 || occupied.{i-1,j} == 0) then begin
        let dx = (if i < ni then g.{i,j}   else 0.0) -.
                 (if i > 0  then g.{i-1,j} else 0.0) in
        x.{i,j} <- x.{i,j} -. dx
      end
  done done;
  for i = 0 to ni-1 do
    for j = 0 to nj do
      if (j == nj || occupied.{i,j} == 0) &
         (j == 0 || occupied.{i,j-1} == 0) then begin
        let dy = (if j < nj then g.{i,j}   else 0.0) -.
                 (if j > 0  then g.{i,j-1} else 0.0) in
        y.{i,j} <- y.{i,j} -. dy
      end
  done done


(* return norm of the grid *)
let norm (g:t) =
  let n = (Array2.dim1 g)*(Array2.dim2 g) in
  let g1 = reshape_1 (genarray_of_array2 g) n in
  let result = ref 0.0 in
  for i = 0 to n-1 do
    result := !result +. g1.{i}*.g1.{i}
  done;
  sqrt !result


(*********************** helper functions for solve_poisson ******************)

type vector = (float, float64_elt, c_layout) Array1.t

let vector_of_grid (g:t) =
  let n = (Array2.dim1 g)*(Array2.dim2 g) in
  reshape_1 (genarray_of_array2 g) n


(* probably better to do this with the BLAS... *)
let dot (x:vector) (y:vector) =
  let n = Array1.dim x in
  let result = ref 0.0 in
  for i = 0 to n-1 do
    result := !result +. x.{i}*.y.{i}
  done;
  !result


(* probably better to do this with the BLAS... *)
let add_scaled (x:vector) scalar (y:vector) =
  let n = Array1.dim x in
  for i = 0 to n-1 do
    x.{i} <- x.{i} +. scalar*.y.{i}
  done


(* probably better to do this with the BLAS... *)
let rescale_add scalar (x:vector) (y:vector) =
  let n = Array1.dim x in
  for i = 0 to n-1 do
    x.{i} <- scalar*.x.{i} +. y.{i}
  done


let apply_laplacian (occupied:Intgrid.t) (xgrid:t) (ygrid:t) =
  let endi = (Array2.dim1 xgrid)-1
  and endj = (Array2.dim2 xgrid)-1 in
  for i = 0 to endi do
    for j = 0 to endj do
      if occupied.{i,j} == 0 then begin
        let xij = xgrid.{i,j} in
        ygrid.{i,j} <- 
          (if i<endi then (if occupied.{i+1,j}==0 then
                             xgrid.{i+1,j}-.xij else 0.0)
           else -.xij)
        +.(if i>0 then (if occupied.{i-1,j} == 0 then
                          xgrid.{i-1,j}-.xij else 0.0)
           else -.xij)
        +.(if j<endj then (if occupied.{i,j+1} == 0 then
                        xgrid.{i,j+1}-.xij else 0.0)
           else -.xij)
        +.(if j>0 then (if occupied.{i,j-1} == 0 then
                          xgrid.{i,j-1}-.xij else 0.0)
           else -.xij)
      end else
        ygrid.{i,j} <- xgrid.{i,j} (* or 0; this keeps operator non-singular *)
  done done


let eval_residual (occupied:Intgrid.t) (x:vector) (xgrid:t) (b:vector)
                  (r:vector) (rgrid:t) =
  apply_laplacian occupied xgrid rgrid;
  let n = Array1.dim r in
  for i = 0 to n-1 do
    r.{i} <- b.{i} -. r.{i}
  done


let apply_preconditioner (occupied:Intgrid.t) (xgrid:t) (ygrid:t) =
  Array2.blit xgrid ygrid   (* identity - the simplest approx *)
  (*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*)


(* the core of the preconditioned conjugate gradient algorithm *)
let rec pcg (occupied:Intgrid.t) (x:vector) (r:vector) (rgrid:t) (p:vector)
            (pgrid:t) (m:vector) (mgrid:t) oldmr oldpr itsleft tol2 =
  apply_laplacian occupied pgrid mgrid;        (* m = A*p *)
  let alpha = oldpr /. dot p m in     (* alpha = (p.r)/(p.A*p) *)
  add_scaled r (-.alpha) m;           (* r = r-alpha*m *)
  add_scaled x alpha p;               (* x = x+alpha*p *)
  if itsleft <= 1 then ()             (* check if ran out of iterations *)
  else begin
    apply_preconditioner occupied rgrid mgrid; (* m = M*r *)
    let newmr = dot m r in
    if newmr <= tol2 then ()          (* check if converged *)
    else begin
      let beta = newmr /. oldmr in    (* beta = (new r.M*r)/(old r.M*r) *)
      rescale_add beta p m;           (* p = m+beta*p *)
      let newpr = dot p r in
      pcg occupied x r rgrid p pgrid m mgrid newmr newpr (itsleft-1) tol2
    end
  end

(*************************** end of helper functions *************************)

(* Poisson solve (homogenous Dirichlet boundaries):
 * Take the given guess in lhs and refine it to be close to the solution
 * of laplacian*lhs = rhs, using the two temporary grids (rhs will be
 * overwritten as well) and according to the given tolerance and maximum
 * number of iterations. We use the undivided, 7-point stencil.
 * Assume Dirichlet boundaries. *)
let solve_poisson (occupied:Intgrid.t) ~(lhs:t) ~(rhs:t) (rgrid:t) (pgrid:t)
                  maxits tol =
  let tol2 = tol *. tol in
  (* for efficiency we will convert the grids to vectors *)
  let x = vector_of_grid lhs
  and r = vector_of_grid rgrid
  and p = vector_of_grid pgrid
  and m = vector_of_grid rhs in
  eval_residual occupied x lhs m r rgrid;           (* r0 = rhs - A*lhs *)
  apply_preconditioner occupied rgrid pgrid;            (* p0 = M*r0 *)
  let oldmr = dot p r in
  if oldmr <= tol2 then ()  (* the initial guess was exact! *)
  else
    (* now iterate. (overwrite rhs - use it as a temporary vector) *)
    pcg occupied x r rgrid p pgrid m rhs oldmr oldmr maxits tol2

